
/*
 * tclsock.c --
 *     This file will mak winsock functions availible to tcl & tk
 *
 */

#include <tcl.h>
#include <errno.h>
#include <time.h>
#include <malloc.h>

#ifndef linux
#define Windows
#endif

#ifdef Windows
#define BLOCK WSAEWOULDBLOCK
#include <winsock.h>
#define SOCKMSG (WM_USER + 1)

static struct WSErr {
	int	err;
	char 	*name;
	} WSerr[] = {
/*
 * All Windows Sockets error constants are biased by WSABASEERR from
 * the "normal"
 */
		{WSABASEERR,"WSABASEERR"},
/*
 * Windows Sockets definitions of regular Microsoft C error constants
 */
		{WSAEINTR ,"Interupted System Call"},
		{WSAEBADF ,"Bad File Number"},
		{WSAEACCES,"Permission Denied"},
		{WSAEFAULT,"Bad Address"},
		{WSAEINVAL,"Invalid Argument"},
		{WSAEMFILE,"Too many open files"},

/*
 * Windows Sockets definitions of regular Berkeley error constants
 */
		{WSAEWOULDBLOCK,"Operation would block"},
		{WSAEINPROGRESS,"Operation now in progress"},
		{WSAEALREADY   ,"Operation already in progress"},
		{WSAENOTSOCK   ,"Socket operation on nonsocket"},
		{WSAEDESTADDRREQ,"Destination address required"},
		{WSAEMSGSIZE   ,"Message too long"},
		{WSAEPROTOTYPE ,"Protocol wrong type for socket"},
		{WSAENOPROTOOPT,"Protocol not availible"},
		{WSAEPROTONOSUPPORT,"Protocol not supported"},
		{WSAESOCKTNOSUPPORT,"Socket type not supported"},
		{WSAEOPNOTSUPP    ,"Operation not supported on socket"},
		{WSAEPFNOSUPPORT ,"Protocol family not supported"},
		{WSAEAFNOSUPPORT,"Address family not supported"},
		{WSAEADDRINUSE  ,"Address already in use"},
		{WSAEADDRNOTAVAIL,"Can't assign requested address"},
		{WSAENETDOWN     ,"Network is down"},
		{WSAENETUNREACH  ,"Network is unreachable"},
		{WSAENETRESET    ,"Network dropped connection on reset"},
		{WSAECONNABORTED ,"Software coused connection abort"},
		{WSAECONNRESET   ,"Connection reset by peer"},
		{WSAENOBUFS      ,"No buffer space availible"},
		{WSAEISCONN      ,"Socket is already connected"},
		{WSAENOTCONN     ,"Socket is not connected"},
		{WSAESHUTDOWN    ,"Can't send after socket shutdown"},
		{WSAETOOMANYREFS ,"Too many references: Can't splice"},
		{WSAETIMEDOUT    ,"Connection timed out"},
		{WSAECONNREFUSED , "Connection refused"},
		{WSAELOOP        , "Too many levels of symbolic links"},
		{WSAENAMETOOLONG ,"File name too long"},
		{WSAEHOSTDOWN    ,"Host is down"},
		{WSAEHOSTUNREACH ,"No route to host"},
		{WSAENOTEMPTY    ,"Directory not empty"},
		{WSAEPROCLIM     ,"Too many processes"},
		{WSAEUSERS       ,"Too many users"},
		{WSAEDQUOT       ,"Disk quota exceeded"},
		{WSAESTALE       ,"Stale NFS file handle"},
		{WSAEREMOTE      ,"Too many levels of remote in path"},
		{WSAEDISCON      ,"Disconnected"},

/*
 * Extended Windows Sockets error constant definitions
 */
		{WSASYSNOTREADY , "Network subsystem is unusable"},
		{WSAVERNOTSUPPORTED,"Winsock DLL cannot support this application"},
		{WSANOTINITIALISED,"Winsock not initialized"},

/* Authoritative Answer: Host not found */
		{WSAHOST_NOT_FOUND,"Host not found"},
/* Non-Authoritative: Host not found, or SERVERFAIL */
		{WSATRY_AGAIN   ,"Nonauthoritve host not found"},
/* Non recoverable errors, FORMERR, REFUSED, NOTIMP */
		{WSANO_RECOVERY ,"Nonrecoverable error"},

/* Valid name, no data record of requested type */
		{WSANO_DATA     ,"Valid name, no data record of requestd type"},
		{ 0 , NULL },

	};

#else
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>

typedef struct sockaddr_in SOCKADDR_IN;
typedef struct sockaddr SOCKADDR;
typedef struct linger LINGER;
typedef int	SOCKET;
typedef char *	LPSTR;

#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#define BLOCK EWOULDBLOCK
#define closesocket close

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE !TRUE
#endif

void sigalarm (int sig )
{
	fprintf(stderr,"Alarm(%d)\n",sig);
}

int WSAGetLastError() 
{
	return errno;
}


#endif

#define MAX_SOCK_READ 2000
static char ubuf[MAX_SOCK_READ * 2];

/*
 * Prototypes for procedures defined later in this file:
 */
int TclSockInit_Var = 0;
static int	TS_debug _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_gmtime _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_getpeername _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_getpeername _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_info _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_listen _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_accept _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_bind _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_cancel _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_close _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_send _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_sendto _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_recv _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_recvfrom _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_sockinit _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_connect _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	TS_socket _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));

static struct prot {
	char 	*pstr;
	int	prot;
	} protocols []= {
#ifdef IPPROTO_IP
		{ "ip",  IPPROTO_IP },
#endif
#ifdef IPPROTO_ICMP
		{ "icmp",IPPROTO_ICMP},
#endif
#ifdef IPPROTO_GGP 
		{ "ggp", IPPROTO_GGP},
#endif
#ifdef IPPROTO_TCP
		{ "tcp", IPPROTO_TCP},
#endif
#ifdef IPPROTO_PUP
		{ "pup", IPPROTO_PUP},
#endif
#ifdef IPPROTO_UDP
		{ "udp", IPPROTO_UDP},
#endif
#ifdef IPPROTO_IDP
		{ "idp", IPPROTO_IDP},
#endif
#ifdef IPPROTO_ND
		{ "nd",  IPPROTO_ND},
#endif
#ifdef IPPROTO_RAW
		{ "raw", IPPROTO_RAW},
#endif
#ifdef IPPROTO_MAX
		{ "max", IPPROTO_MAX},
#endif
		{ NULL, 0 }
	};

static char * cvtupper ( char * s )
{
	int i;

	for(i = (strlen(s) -1); i >= 0; i--)
		{
		char c;
		c = (char )toupper(s[i]);
		s[i] = c;
		}
	
	return s;
}	

static char * cvtlower ( char * s )
{
	int i;

	for(i = (strlen(s) -1); i >= 0; i--)
		{
		char c;
		c = (char )tolower(s[i]);
		s[i] = c;
		}
	
	return s;
}	

static int prot(char *p )
{
	int i;

	for(i=0; protocols[i].pstr; i++ )
		if ( !strcmp(cvtlower(p),protocols[i].pstr) )
			return protocols[i].prot;
	return SOCKET_ERROR;
}

static char * pstr(int p )
{
	int i;

	for(i=0; protocols[i].pstr; i++ )
		if ( p == protocols[i].prot )
			return protocols[i].pstr;
	return NULL;
}
	

static void TS_cleanup (void)
{
#ifdef Windows
	if ( TclSockInit_Var )
		WSACleanup();
#endif
}

static char *TS_WSerror(void)
{


#ifdef Windows
	int i,err;

	for( err = WSAGetLastError(),i = 0; WSerr[i].name; i++ )
		if ( WSerr[i].err == err )
			return WSerr[i].name;
	return "Unknown Winsock error";
#else

	static char b[20];
	sprintf(b,"%d",errno);
	return b;
#endif
	
}
/**********************************************
*  Get the host address
*  If the host starts with a digit, to convert an inet addr
*  Otherwise use gethost by name
**********************************************/
static int gethost(char *host, SOCKADDR_IN *name)
{
	struct hostent *hp;

	if ( isdigit(*host) )
		name->sin_addr.s_addr = inet_addr(host);
	else
		{
		if ( (hp = gethostbyname(host)) == NULL )
			return TCL_ERROR;
		memcpy((char *)&name->sin_addr, hp->h_addr, hp->h_length);
		}
	return TCL_OK;

}

static int getserv(char *serv, char *protocol, SOCKADDR_IN *name )
{
	struct servent *sp;

	if ( isdigit(*serv) )
		name->sin_port = htons((short)atoi(serv));
	else
	{
	if ( (sp = getservbyname(serv,protocol)) == NULL)
		if ( (sp=getservbyport(atoi(serv),protocol)) == NULL)
			return TCL_ERROR;
	name->sin_port   = sp->s_port;
	}

	return TCL_OK;
}

#define MAXSOCK	20

struct TclSocket {
	int 		stat;
#define SOCKCLOSED	0
#define SOCKOPEN	1
#define SOCKBIND	2
	SOCKET		s;
	int 		protocol;  /* ip, tcp, udp ect. */
	int		type;      /* stream or datagram */
	char *		host;
	int 		async;
#ifdef Windows
	HWND		w;
#endif
	SOCKADDR_IN	addr;
	int		last_error;
	int		bind;
	char 		*buf;
	int 		count;
	int 		cancel;
	int		inprogress;
	} TclSocket[MAXSOCK+1];
#define NAME 	TclSocket[s].addr
#define BUFFER	TclSocket[s].buf

static int dbflag = 0;

#define DEBUG	\
		if (dbflag) \
			{  \
			fprintf(stderr,"DB: %s (%d) socket(%d) f(%s)\n", \
				__FILE__, __LINE__, s ,argv[0] ); \
			fflush(stderr); \
			} 
			

/************************************************************************
*  This macro will check the first argument to insure it is a valid
*       socket.
************************************************************************/
#define CHECKARGS(Min,Max,Msg) \
	if ( argc < Min || argc > Max ) \
		{ \
		sprintf(interp->result , Msg ); \
		return TCL_ERROR; \
		} \
	if ( strncmp("TclSocket",argv[1],9) != 0 || !isdigit(argv[1][9])) \
		{ \
		sprintf(interp->result , "Invalid socket in %s (%s)", \
			argv[0], argv[1]); \
		return TCL_ERROR; \
		} \
	if ( (s = atoi(argv[1]+9)) > MAXSOCK  \
		|| s < 0  \
		|| TclSocket[s].stat != SOCKOPEN ) \
		{ \
		sprintf(ubuf , "Invalid socket stat in %s (%s) stat = (%d) s = (%d)", \
			argv[0],argv[1],TclSocket[s].stat,s) ; \
		interp->result = ubuf; \
		return TCL_ERROR; \
		}  \
	if (TclSocket[s].inprogress && strcmp(argv[0],"cancelsocket") ) \
		{ \
		sprintf(ubuf , "socket operation in progress %s (%s) stat = (%d) s = (%d)", \
			argv[0],argv[1],TclSocket[s].stat,s) ; \
		interp->result = ubuf; \
		return TCL_ERROR; \
		}  \
	DEBUG 

#ifndef Windows

#ifndef FD_READ
#define FD_READ 1
#endif
#ifndef FD_WRITE
#define FD_WRITE 2
#endif

/***********************************************************************************
*  In UNIX, wait until the socket is ready for the operation.
*	every two seconds do an update to let TK do other stuff.
*    	One of the other things TK may do is call cancel_socket.  If so
*       return 1.
***********************************************************************************/
static int waitfor( Tcl_Interp *interp, struct TclSocket *s, int action ) 
{
	fd_set rfds,wfds;
	struct timeval tv;
        int done = 0;

	tv.tv_sec = 2;  /*  Only wait 2 seconds then check to see if the op has been canceled */
	tv.tv_usec = 0;

	s->inprogress ++;
	while( !done ) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		switch (action) {
			case FD_READ: FD_SET(s->s,&rfds);break;
			case FD_WRITE: FD_SET(s->s,&wfds);break;
		}

		done = select( s->s+1, &rfds, &wfds, NULL, &tv);
		if ( !done )
			{
			Tcl_Eval(interp,"update");
			done = s->cancel;
			}
		else
			{
			/* Do a sanity Check */
			if( !FD_ISSET(s->s,&rfds) && !FD_ISSET(s->s,&wfds) )
				{
				fprintf(stderr,"Error in waitfor.  no socket read after select!\n");
				}
			}
	}
	s->inprogress--;
	return s->cancel;	
}

#endif

static int TS_info(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;

	CHECKARGS(2,3, "Wrong # of args (sockinfo socket v[erbose])")


	if( argc == 3 )
		sprintf(ubuf, "ipaddr(%s) port(%d) bind(%d) protocol(%d:%s) type(%s) family(%s)",
			inet_ntoa(TclSocket[s].addr.sin_addr),
			ntohs(TclSocket[s].addr.sin_port),
			TclSocket[s].bind,
			TclSocket[s].protocol,
			pstr(TclSocket[s].protocol),
			TclSocket[s].type == 
				SOCK_DGRAM ? "SOCK_DGRAM" : "SOCK_STREAM",
			TclSocket[s].addr.sin_family == 
				AF_INET ? "AF_INET" : "UNKNOWN FAMILY"
			);
	
	else
	sprintf(ubuf, "{%s} {%d}",
			inet_ntoa(TclSocket[s].addr.sin_addr),
			ntohs(TclSocket[s].addr.sin_port)
			);
	interp->result = ubuf;
	return TCL_OK;

}

static int TS_listen(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int backlog;

	CHECKARGS(2,3, "Wrong # of args (listen socket [bacllog])")

	if ( !TclSocket[s].bind )
		{
		sprintf(ubuf,"%s is not bound to a port",argv[1]);
		interp->result = ubuf;
		return TCL_ERROR;
		}

	if ( argc == 3)
		backlog = atoi(argv[2]);
	else
		backlog = 5;

	if ( listen(TclSocket[s].s, backlog) == SOCKET_ERROR)
		{
		sprintf(ubuf,"%s Can't listen at %s (%s)",
				argv[1], argv[2], TS_WSerror());
		interp->result = ubuf;
		return TCL_ERROR;
		}

	interp->result = "OK";
	return TCL_OK;
}



static int TS_accept(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s,s2;
	int done;
	SOCKET	ns;

	CHECKARGS(2,3, "Wrong # of args (accept socket [host/address])")

	done = 0;

	if ( argc == 3)
		{
		if( gethost(argv[2],&TclSocket[s].addr) == TCL_ERROR)
			{
			sprintf(ubuf,"Can't get address for %s",argv[2]);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_ACCEPT ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "Can't select for accept %s <%s>",
				argv[1], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif


while ( !done )
	{
	int err;
	
#ifdef Windows
	TclSocket[s].async = 1;
	while ( TclSocket[s].async && !done)
		{
		MSG msg;
		PeekMessage(&msg,NULL,0,0,PM_REMOVE);
		if (msg.hwnd==TclSocket[s].w )
			TclSocket[s].async = 0;
		else 
			{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			}
		}
#endif
	
	if ((ns=accept(TclSocket[s].s,NULL,NULL)) == INVALID_SOCKET)
		{
		if ( (err = WSAGetLastError()) != BLOCK) 
			{
			sprintf(ubuf , "Can't accept %s (%d)",argv[1],err);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}
	else
		done = 1;
	}
	
	/* Find the next open slot in our list */
	for( s2 = 0 ; s2<MAXSOCK; s2++)
		if( TclSocket[s2].stat == SOCKCLOSED)
			break;
	if( s2 == MAXSOCK )
		{
		closesocket(ns);
		interp->result = "No sockect availible!";
		return TCL_ERROR;
		}

	TclSocket[s2].s = ns;

	TclSocket[s2].stat = SOCKOPEN;

	/* These should be the same as the original */
	TclSocket[s2].type     = TclSocket[s].type    ;
	TclSocket[s2].addr.sin_family   = TclSocket[s].addr.sin_family  ;
	TclSocket[s2].protocol = TclSocket[s].protocol;

#ifdef Windows
	if ( (TclSocket[s2].w = CreateWindow("STATIC",interp->result, 
				WS_OVERLAPPED, 0,0,0,0,NULL,NULL,NULL,NULL))==0)
		{
		sprintf(interp->result , "Can't create hidden window %s",s);
		closesocket(TclSocket[s2].s);

		TclSocket[s2].stat = SOCKCLOSED;
		return TCL_ERROR;
		}
#endif

	sprintf(interp->result , "TclSocket%d",s2);
	return TCL_OK;
}


static int TS_bind(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int err;
	char * pname;

	CHECKARGS(3,4, "Wrong # of args (bind socket port [host/address])")
	

if ( TclSocket[s].bind )
	{
	sprintf(interp->result , "%s is already bound!",argv[1]);
	return TCL_ERROR;
	}
	NAME.sin_family = TclSocket[s].addr.sin_family;

	if ( (pname = pstr(TclSocket[s].protocol)) == NULL)
		{
		sprintf(ubuf,"%s has an invalid protocol (%d)",
				TclSocket[s].protocol);
		interp->result = ubuf;
		return TCL_ERROR;
		}

	/* Get the port number */
	if ( isdigit(*argv[2]) )
		NAME.sin_port	= htons((short)atoi(argv[2]));
	else 
		{
		if ( getserv(argv[2], pname,&NAME) == SOCKET_ERROR)
			{
			sprintf(interp->result,"Can't get service %s , %s",
						argv[2],pname);
			interp->result = ubuf;
				return TCL_ERROR;
			}
		}

	if ( argc == 4 )
	{
	if ( isdigit(*argv[3]) )
		NAME.sin_addr.s_addr = inet_addr(argv[3]);
	else
		{
		if ( gethost(argv[2],&NAME) == SOCKET_ERROR)
			{
			sprintf(interp->result,"Can't get addr for %s",argv[3]);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}

	}
	else
		NAME.sin_addr.s_addr = htonl(INADDR_ANY);

	
	if( (err = bind( TclSocket[s].s,
		(struct sockaddr *)&NAME,sizeof(NAME)))== SOCKET_ERROR)
		{
		sprintf(ubuf,"Can't bind %s at %s",argv[2],err);
		interp->result = ubuf;
		return TCL_ERROR;
		}

	TclSocket[s].bind = TRUE;
	interp->result = "OK";
	return TCL_OK;
}

/*******************************************************************************************
*  Cancel a socket operation.  See waitfor!
*******************************************************************************************/
static int TS_cancel(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;

	CHECKARGS(2,2, "Wrong # of args (cancelsocket socket)")

	TclSocket[s].cancel = 1;
	return TCL_OK;

}

static int TS_close(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int err;
	LINGER ling;

	CHECKARGS(2,2, "Wrong # of args (closesocket socket)")

	ling.l_onoff = 0;
	ling.l_linger = 0;
setsockopt(TclSocket[s].s,SOL_SOCKET,SO_LINGER,(LPSTR)&ling,sizeof(ling));	

	if ( (err = closesocket(TclSocket[s].s) ) == SOCKET_ERROR)
		{
		sprintf(ubuf,"Can't close %s (%d)",argv[1],err);
		interp->result = ubuf;
		return TCL_ERROR;
		}
	free(BUFFER);

	TclSocket[s].stat = SOCKCLOSED;
	TclSocket[s].bind = FALSE;
#ifdef Windows
	DestroyWindow(TclSocket[s].w);
#endif
	interp->result = "OK";
	return TCL_OK;
}

static int TS_sendto(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int done;
	int err;
	int tosend;
	int count;
	int block;
	char *outbuf ;

	CHECKARGS(4,6, "Wrong # of args (sendto socket host port [size] ValueToSend)")
	*ubuf = '\0';
	*BUFFER = '\0';
	done = count = 0;

	if ( argc == 6 )
		{
		outbuf = argv[5];
		tosend = atoi(argv[4]);
		}
	else
		{
		outbuf = argv[4];
		tosend = strlen(outbuf);
		}
		

	/*************************************************
	*  Getthe address of the sendee.
	*************************************************/
	if ( gethost(argv[2],&NAME) != TCL_OK)
		{
		sprintf(ubuf,"Can't find address of %s",argv[2]);
		interp->result = ubuf;
		return TCL_ERROR;
		}
	if ( getserv(argv[3],pstr(TclSocket[s].protocol),&NAME) != TCL_OK)
		{
		sprintf(ubuf,"Can't find port of %s (%s)",argv[3],
				pstr(TclSocket[s].protocol));
		interp->result = ubuf;
		return TCL_ERROR;
		}


		
	block = 0;

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_WRITE ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "Can't select for send %s <%s>",
				argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif

while ( !done  && tosend)
	{
	int sent ;

#ifndef Windows
	/* alarm(2); */
#endif
	
	if ((sent=sendto(TclSocket[s].s,outbuf,tosend,0,(SOCKADDR*)&NAME,sizeof(NAME))) == SOCKET_ERROR)
		{
		if ( (err = WSAGetLastError()) != BLOCK) 
			{
			sprintf(ubuf , "sendto %d err = (%d)",count,err);
			interp->result = ubuf;
			return TCL_ERROR;
			}

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_WRITE ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "sendto %d XCan't select for write %s <%s>",
				count, argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
		TclSocket[s].async = 1;
		while ( TclSocket[s].async && !done)
			{
			MSG msg;
			PeekMessage(&msg,NULL,0,0,PM_REMOVE);
			if (msg.hwnd==TclSocket[s].w )
				{
				TclSocket[s].async = 0;
				}

			else 
				{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
				}
			}
#endif
		}
	else
		{
		count+= sent;
		outbuf+= sent;
		tosend -= sent;
		}
	}

	/* Turn it off */
#ifdef Windows
	WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w, SOCKMSG, 0 );
#else
	/* alarm(0); */ 
#endif
	sprintf(ubuf,"%d SendTo OK <%s> block %d s = %d",count,BUFFER,block,s);
	interp->result = ubuf;
	return TCL_OK;
}
		

/***************************************************************
*  send scoket [size] val
***************************************************************/
static int TS_send(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int done;
	int err;
	int tosend;
	int count;
	int block;
	char *outbuf ;

	CHECKARGS(3,4, "Wrong # of args (send socket [size] ValueToSend)")
	*ubuf = '\0';
	*BUFFER = '\0';
	done = count = 0;

	if ( argc == 4 )
		{
		outbuf = argv[3];
		tosend = atoi(argv[2]);
		}
	else
		{
		outbuf = argv[2];
		tosend = strlen(outbuf);
		}
		

	block = 0;

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_WRITE ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "Can't select for send %s <%s>",
				argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif


while ( !done  && tosend)
	{
	int sent ;
	
#ifndef Windows
	/* alarm(2); */
#endif
	if ((sent=send(TclSocket[s].s,outbuf,tosend,0)) == SOCKET_ERROR)
		{
		if ( (err = WSAGetLastError()) != BLOCK) 
			{
			sprintf(ubuf , "%d err = (%d)",count,err);
			interp->result = ubuf;
			return TCL_ERROR;
			}
#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_WRITE ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "%d XCan't select for write %s <%s>",
				count, argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
		TclSocket[s].async = 1;
		while ( TclSocket[s].async && !done)
			{
			MSG msg;
			PeekMessage(&msg,NULL,0,0,PM_REMOVE);
			if (msg.hwnd==TclSocket[s].w )
				{
				TclSocket[s].async = 0;
				}

			else 
				{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
				if ( block ++ > 20000 ) 
					done = 1;
				}
			}
#endif
		}
	else
		{
		count+= sent;
		outbuf+= sent;
		tosend -= sent;
		}
	}

#ifdef Windows
	/* Turn it off */
	WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w, SOCKMSG, 0 );
#else
	/* alarm(0); */
#endif
	sprintf(ubuf,"%d Write OK <%s> block %d s = %d",count,BUFFER,block,s);
	interp->result = ubuf;
	return TCL_OK;
}
		
static int TS_recvfrom(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int done;
	int err;
	int max_read;
	int count;
	int addrlen;


	CHECKARGS(2,3, "Wrong # of args (read socket max)")
	*ubuf = '\0';
	*BUFFER = '\0';
	done = count = 0;


	if( argc == 3 ) 
		{
		if( (max_read = atoi(argv[2])) > MAX_SOCK_READ-2)
			max_read = MAX_SOCK_READ-2;
		}
	else
		max_read = MAX_SOCK_READ-2;

	if ( !TclSocket[s].bind )
		{
		sprintf(ubuf,"%s is not bound to a port",argv[1]);
		interp->result = ubuf;
		return TCL_ERROR;
		}

	TclSocket[s].addr.sin_addr.s_addr = htonl(INADDR_ANY);

	addrlen = sizeof(TclSocket[s].addr);

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_READ ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "Can't select for read %s <%s>",
				argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif

	sprintf(ubuf,"ok, max_read = %d sizeof %d struct %d",
		max_read,addrlen,sizeof(SOCKADDR_IN));
	interp->result = ubuf;

while( !done )
{

#ifndef Windows
	/* alarm(2);  */ /* Set an interupt in two seconds */
#endif
if ( (count = recvfrom(TclSocket[s].s, BUFFER, max_read, 0 ,
		(SOCKADDR*)&TclSocket[s].addr,&addrlen)) == SOCKET_ERROR)
		{
		if ( (err = WSAGetLastError()) != BLOCK) 
			{
			sprintf(ubuf , "2 Can't recv %s err=(%d) <%s>",
				argv[2], err, BUFFER);
			interp->result = ubuf;
			return TCL_ERROR;
			}
#ifdef Windows
		TclSocket[s].async = 1;
		while ( TclSocket[s].async )
			{
			MSG msg;
			PeekMessage(&msg,NULL,0,0,PM_REMOVE);
			if (msg.hwnd==TclSocket[s].w )
				{
				TclSocket[s].async = 0;
				}

			else 
				{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
				}
			}
#endif
		}
	else
		done = 1;
}

	/* Turn it off */
#ifdef Windows
	WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w, SOCKMSG, 0 ) ;
#else
	/* alarm(0);  */ /* Set an interupt in two seconds */
#endif
	BUFFER[count] = '\0';
	interp->result = BUFFER;
	return TCL_OK;
}
		

/***********************************************************************
*  read socket [size]
*  Read data from a socket, it size is not specified, read until
*       a newline is read.
***********************************************************************/
static int TS_recv(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int done;
	int err;
	int max_read;
	int count;
	int block;
	int sznl;
	char *newline;

	CHECKARGS(2,4, "Wrong # of args (read socket [NoBytes | -newline char])")
	*ubuf = '\0';
	*BUFFER = '\0';
	done = count = 0;

	newline = "\r\n";
	max_read = MAX_SOCK_READ-2;

	if ( argc == 4 )
		{
		if( !strcmp(argv[3],"-newline") )
			newline = argv[4];
		else
			{
			sprintf(ubuf, "Invalid arg to recv (%s)",argv[3]);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}

	if ( argc == 3 )
		{
		if ( (max_read = atoi(argv[2])) >= (MAX_SOCK_READ-2))
			{
			max_read = MAX_SOCK_READ-2;
			newline = NULL;
			}
		else 
			max_read = MAX_SOCK_READ-2;
		}

	sznl = strlen(newline);
	block = 0;
#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_READ ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "Can't select for read %s <%s>",
				argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif

while ( !done  && count < max_read)
	{
	unsigned char	ui;
	
#ifndef Windows
	if ( waitfor(interp,&TclSocket[s], FD_READ) )
		goto out;
#endif
	if ( recv(TclSocket[s].s, (char *)&ui, 1, 0 ) == SOCKET_ERROR)
		{
		if ( (err = WSAGetLastError()) != BLOCK) 
			{
			/********************************************
			*  If some data was retrived send it back.
			*  the next call to read will probably couse
			*  the same error.
			********************************************/
			if ( count )
				{
				BUFFER[count] = '\0';
				interp->result = BUFFER;
				return TCL_OK;
				}
			else
				{
				sprintf(ubuf , "2 Can't recv %s err=(%d) <%s>",
					argv[2], err, BUFFER);
				interp->result = ubuf;
				return TCL_ERROR;
				}
			}
#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
		SOCKMSG, FD_READ ) == SOCKET_ERROR)
		{
		/********************************************
		*  If some data was retrived send it back.
		*  the next call to read will probably couse
		*  the same error.
		********************************************/
		if ( count )
			{
			BUFFER[count] = '\0';
			interp->result = BUFFER;
			return TCL_OK;
			}
		else
			{
		sprintf(ubuf , "XCan't select for read %s <%s>",
				argv[2], BUFFER);
		interp->result = ubuf;
		return TCL_ERROR;
			}
		}
		TclSocket[s].async = 1;
		while ( TclSocket[s].async && !done)
			{
			MSG msg;
			PeekMessage(&msg,NULL,0,0,PM_REMOVE);
			if (msg.hwnd==TclSocket[s].w )
				{
				TclSocket[s].async = 0;
				}

			else 
				{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
				if ( block ++ > 20000 ) 
					done = 1;
				}
			}
#endif
		}
	else
		{
		BUFFER[count++] = (char)ui;
		BUFFER[count] = '\0';
		if (!strcmp(&(BUFFER[count-sznl]),newline ) )
			{
			done = 1;
			count -= sznl;
			}
		}
	}

	/* Turn it off */
#ifdef Windows
	WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w, SOCKMSG, 0 ) ;
#endif
out:
	TclSocket[s].cancel = 0 ;
	BUFFER[count] = '\0';
	interp->result = BUFFER;
	return TCL_OK;
}
		
/*******************************************************************************
*  This function returns the current system dat eand time.
*  This is here to provide a consitent way of getting this information,
*  in both MSDOS & UNIX.
*  The time is returned in the form "YYMMDD HHMMSS" The time returned
*  is GMT.
*******************************************************************************/
static int TS_gmtime(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	time_t t;
	struct tm *tm;
	
	time(&t);
	tm = gmtime(&t);
	
	sprintf(ubuf,"%-2.2d%-2.2d%-2.2d %-2.2d%-2.2d%-2.2d",
		tm->tm_year, tm->tm_mon+1, tm->tm_mday,
		tm->tm_hour+1, tm->tm_min+1, tm->tm_sec+1
		);

	interp->result = ubuf;
	return TCL_OK;


}

/*************************************************************
*   Just set a debug flag
*************************************************************/
static int TS_debug(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{

	dbflag = atoi(argv[1]);
	interp->result = "";
	return TCL_OK;


}

/**********************************************************************
* Connect socket remoteHost service
*      socket = an open socket from a call to socket;
*      remoteHost = hostname or ip address
*      service    = service name or port number
***********************************************************************/
static int TS_connect(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int protocol;
#ifdef Windows
	MSG msg;
#endif
	int count = 0;

CHECKARGS(4,5,"Wrong # of args (connect socket host service [protocol])");


	if ( gethost(argv[2],&NAME) == SOCKET_ERROR )
		{
		sprintf(interp->result ,"Can't get addr for %s",argv[2]);
		return TCL_ERROR;
		}
	

	if ( argc == 5 )
		{
		if ( (protocol = prot(argv[4])) == SOCKET_ERROR)
			{
			sprintf(interp->result,"Invalid protocol %s",
				argv[3]);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}
	else 
		protocol = IPPROTO_TCP;

	if ( getserv(argv[3],pstr(protocol),&NAME) == SOCKET_ERROR)
		{
		sprintf(interp->result,"Can't get service %s , %s",
				argv[3],pstr(protocol));
		interp->result = ubuf;
		return TCL_ERROR;
		}

	NAME.sin_family = AF_INET;
		
	*BUFFER = '\0';	

	TclSocket[s].async = 1;

#ifdef Windows
	if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
			SOCKMSG, FD_CONNECT ) == SOCKET_ERROR)
		{
		sprintf(ubuf , "2 Can't select to %s at %s (%s) <%s> count %d", 
				argv[2], argv[3] ,TS_WSerror(),BUFFER,count);
		interp->result = ubuf;
		return TCL_ERROR;
		}
#endif

	/* Wait for the message to be processed */
	*BUFFER = '\0';
	if ( connect(TclSocket[s].s, (struct sockaddr *)&NAME, 
		sizeof(struct sockaddr)) == SOCKET_ERROR )
		{
		count++;
			if ( WSAGetLastError() != BLOCK)
			{
			sprintf(ubuf , 
				"Can't connect to %s at %s (%s) <%s> count %d", 
				argv[2], argv[3] ,TS_WSerror(),BUFFER);
			interp->result = ubuf;
			return TCL_ERROR;
			}
		}
#ifdef Windows
	while ( TclSocket[s].async )
		{
		PeekMessage(&msg,NULL,0,0,PM_REMOVE);
		if ( msg.hwnd == TclSocket[s].w && msg.lParam == FD_CONNECT)
			{
			TclSocket[s].async = 0;
			if ( WSAAsyncSelect(TclSocket[s].s,TclSocket[s].w,
					SOCKMSG, 0 ) == SOCKET_ERROR)
				{
				sprintf(ubuf , "Can't turn off async %s at %s (%s) <%s> count %d lparm=%ld CONNECT=%d", 
				argv[2], argv[3] ,TS_WSerror(),BUFFER,count,
				msg.lParam,FD_CONNECT);
				interp->result = ubuf;
				return TCL_ERROR;
				}
			sprintf(BUFFER,"lparam = %d Connect %d",msg.lParam,FD_CONNECT);
			}
		else
			{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			if ( !TclSocket[s].async && *BUFFER )
			{
				sprintf(ubuf,"Activity in async <%s>",BUFFER);
				interp->result = ubuf;
				return TCL_ERROR;
			}
			}
		}
#endif


	sprintf(interp->result , "Connect OK (%d) <%s>",s,BUFFER);
	return TCL_OK;

}


static int TS_getsockname(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int alen;
	SOCKADDR_IN addr;

	CHECKARGS(2,2, "Wrong # of args (getsockname socket)")

	alen = sizeof(addr);

	if ( getsockname(TclSocket[s].s, (SOCKADDR*)&addr, &alen)== SOCKET_ERROR)
		{
		sprintf(ubuf,"Error %s",TS_WSerror());
		interp->result = ubuf;
		return TCL_ERROR;
		}


	sprintf(ubuf, "{%s} {%d}",
			inet_ntoa(addr.sin_addr),
			ntohs(addr.sin_port)
			);
	interp->result = ubuf;
	return TCL_OK;
	

}

static int TS_getpeername(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int alen;
	SOCKADDR_IN addr;

	CHECKARGS(2,2, "Wrong # of args (getpeername socket)")

	alen = sizeof(addr);

	if ( getpeername(TclSocket[s].s, (SOCKADDR*)&addr, &alen)== SOCKET_ERROR)
		{
		sprintf(ubuf,"Error %s",TS_WSerror());
		interp->result = ubuf;
		return TCL_ERROR;
		}


	sprintf(ubuf, "{%s} {%d}",
			inet_ntoa(addr.sin_addr),
			ntohs(addr.sin_port)
			);
	interp->result = ubuf;
	return TCL_OK;
}


/* Initialize a socket */
static int TS_socket(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int s;
	int protocol, type;
	u_short family;

	if ( argc < 1 || argc > 3 )
		{
		interp->result = "Wrong number of args [type] [protocol]";
		return TCL_ERROR;
		}


	/* Set the defaults */
	family = AF_INET;
	type = SOCK_STREAM;
	protocol = IPPROTO_TCP;
	
	if ( argc >= 2)
		{
		if ( !strcmp(cvtupper(argv[1]),"DATAGRAM") 
			||  !strcmp(cvtupper(argv[1]),"DGRAM") )
			{
			type = SOCK_DGRAM;
			protocol = IPPROTO_IP;
			}
		else if ( !strcmp(cvtupper(argv[1]),"STREAM") )
			{
			type = SOCK_STREAM;
			protocol = IPPROTO_TCP;
			}
		else {
			interp->result = "Invalid socket type (datagram or stream)";
			return TCL_ERROR;
			}
		}
		
	if ( argc == 3)
		{
		if ( (protocol = prot(argv[2])) ==  SOCKET_ERROR  )
			 {
			interp->result = 
			"Invalid protocol (ip,icmp,ggp,tcp,pup,udp,idp,nd,raw,max)";
			return TCL_ERROR;
			}
		}
			
	if ( !TclSockInit_Var )
		{
		int err;
		if ( (err = TS_sockinit(dummy,interp,argc,argv)) != TCL_OK )
			return err;
		}

	for( s = 0 ; s<MAXSOCK; s++)
		if( TclSocket[s].stat == SOCKCLOSED)
			break;
	if( s == MAXSOCK )
		{
		interp->result = "No sockect availible!";
		return TCL_ERROR;
		}
	TclSocket[s].s = socket(family, type, protocol);
	if ( TclSocket[s].s == INVALID_SOCKET )
		{
		sprintf(interp->result , "Can't create socket (%s)",
			TS_WSerror());
		return TCL_ERROR;
		}
	else 
		{
		u_long one = 1;

		/*  Put the socket in non blocking mode just in case */
#ifdef Windows
		ioctlsocket( TclSocket[s].s , FIONBIO, &one );  
#endif
		}

	sprintf(interp->result , "TclSocket%d",s);
#ifdef Windows
	if ( (TclSocket[s].w = CreateWindow("STATIC",interp->result, 
				WS_OVERLAPPED, 0,0,0,0,NULL,NULL,NULL,NULL))==0)
		{
		sprintf(interp->result , "Can't create hidden window %s",s);
		closesocket(TclSocket[s].s);

		TclSocket[s].stat = SOCKCLOSED;
		return TCL_ERROR;
		}
#endif
	/* All is cool hear */

	TclSocket[s].stat = SOCKOPEN;
	TclSocket[s].type = type;
	TclSocket[s].addr.sin_family = family;
	TclSocket[s].protocol = protocol;
	TclSocket[s].buf = malloc(MAX_SOCK_READ+1);
	TclSocket[s].cancel = 0 ;
	TclSocket[s].inprogress = 0 ;
	*TclSocket[s].buf = '\0';
	return TCL_OK;
}

static int TS_sockinit(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	int errcode;

#ifdef Windows
	WSADATA wsdata;
	WORD req;
	req = MAKEWORD(1,1);
#endif

	interp->result = "OK";
	if ( !TclSockInit_Var )
		{
		int i;
#ifdef Windows
		if ( (errcode = WSAStartup(req, &wsdata)) != 0 )
			{
			interp->result = "Can't init winsock";
			TclSockInit_Var = errcode;
			return TCL_ERROR;
			}
#endif
		TclSockInit_Var = 1;	
		for( i=0; i< MAXSOCK; i++)
		{

			TclSocket[i].stat = SOCKCLOSED;
		}

#ifdef Windows
	sprintf(ubuf , "winsock init OK! ver = %d,%d. Desc %s.",
			LOBYTE(wsdata.wVersion),
			HIBYTE(wsdata.wVersion),
			wsdata.szDescription);
		interp->result = ubuf;
#endif
		}


	return TCL_OK;
}


/*
/*
 *----------------------------------------------------------------------
 *
 * Socket_Init -- the library initialization file for the extension
 *
 */


int 
#ifdef Windows
_export
#endif 
Socket_Init(interp)
	 Tcl_Interp *interp;		/* Interpreter in which the package is
					 * to be made available. */
{

	 Tcl_CreateCommand(interp, "connect", TS_connect, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "sockinit", TS_sockinit, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "socket", TS_socket, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "recv", TS_recv, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "recvfrom", TS_recvfrom, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "socksend", TS_send, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "sendto", TS_sendto, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "cancelsocket", TS_cancel, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "closesocket", TS_close, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "sockbind", TS_bind, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "accept", TS_accept, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "listen", TS_listen, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "sockinfo", TS_info, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "getpeername",TS_getpeername, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "getsockname",TS_getsockname, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "gmtime",TS_gmtime, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "sockdebug",TS_debug, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	atexit(TS_cleanup);
	return TCL_OK;
}

